
/* rms deviation calculations */

#include <stdio.h>
#include <stdlib.h>	/* contient malloc sur HP */
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

/* -------------------- */
/* parametres		*/
/* -------------------- */
#define DIM3  3
#define cos45 0.707106781
#define sin45 0.707106781
#define precise 0.001	/* default = 10-3 degree */

#define RADIANS 0.017453292
/* -------------------------------------------------------------------- */
/* exported variables */
/* -------------------------------------------------------------------- */
double rms_tolerance = precise*RADIANS;       
double rms_no_rot = precise*RADIANS * 0.25;
double rms_brm[3][3];                   /* matrix of optimal rotation */

/* -------------------------------------------------------------------- */
/*  initializes matrix m as the identity matrix				*/
/* -------------------------------------------------------------------- */
void coo_unit_matrix(double m[3][3])
{
  m[0][0] = m[1][1] = m[2][2] = 1.0;
  m[1][0] = m[0][1] = m[1][2] = 0.0;
  m[2][0] = m[2][1] = m[0][2] = 0.0;
}
/* -------------------------------------------------------------------- */
/*  initializes matrix m as a zero matrix:				*/
/* -------------------------------------------------------------------- */
void coo_zero_matrix(double m[3][3])
{
  m[0][0] = m[0][1] = m[0][2] = 0.0;
  m[1][0] = m[1][1] = m[1][2] = 0.0;
  m[2][0] = m[2][1] = m[2][2] = 0.0;
}
/* -------------------------------------------------------------------- */
/* transposes m */
/* -------------------------------------------------------------------- */
void coo_transpose(double m[3][3])
{
  double sw;

  sw = m[0][1]; m[0][1] = m[1][0]; m[1][0] = sw;
  sw = m[0][2]; m[0][2] = m[2][0]; m[2][0] = sw;
  sw = m[1][2]; m[1][2] = m[2][1]; m[2][1] = sw;
}
/* -------------------------------------------------------------------- */
/* calculates matrix product c = a x b  and returns result in b */
/* -------------------------------------------------------------------- */
void coo_mat_mul(double a[3][3], double b[3][3], double c[3][3])
{
  int 	i, j, k;

  for (i = 0; i < DIM3; i++) 
	for(j = 0; j < DIM3; j++) 
		c[i][j] = a[i][0]*b[0][j]+a[i][1]*b[1][j]+a[i][2]*b[2][j];
}
/* -------------------------------------------------------------------- */
/* rotates the n coordinates coo by rotation matrix mat. */
/* -------------------------------------------------------------------- */
void coo_rot_coo(double **coin, int n, double mat[3][3])
{
  int 	i, j, k;
  double xyz[3];

  for (i = 0; i < n; i++) {
     	for (j = 0; j< DIM3; j++)
	   	xyz[j] = coin[i][j];
     	for (j = 0; j< DIM3; j++)
    	coin[i][j] =  mat[j][0] * xyz[0] + mat[j][1] * xyz[1] + mat[j][2] * xyz[2];
  }
}
/* -------------------------------------------------------------------- */
/* calculates the square of the rms distance of s1 and s2 */
/* -------------------------------------------------------------------- */
double rms_error(double **s1, double **s2, double *cn1, double *cn2, int len)
{
  	double 	rms = 0.0, d, xyz[2][DIM3];
  	int   	i, j, k;

  for (i = 0; i < len; i++) {
    	for (j = 0; j < DIM3; j++){
    		xyz[0][j] = s1[i][j] - cn1[j];
    		xyz[1][j] = s2[i][j] - cn2[j];
	}
    	for (j = 0; j < DIM3; j++){
    		d =  xyz[0][j];
    		for (k = 0; k < DIM3; k++)
    			d -= (rms_brm[j][k]*xyz[1][k]);
		rms += d * d;
	}
  }
  return rms /= len;      /* returns squared rms error */
}
/* -------------------------------------------------------------------- */
/* calculates mixed inertial (cross) tensor of two objects s1 and s2. */
/* -------------------------------------------------------------------- */
void rms_cross_tensor(double **s1, double **s2, double tns[3][3], double *cn1, double *cn2, int len)
{
  double xyz[2][3];
  int   j, k, i;

  coo_zero_matrix(tns);

  for (k = 0; k < len; k++) {
    	for (j = 0; j< DIM3; j++){
    		xyz[0][j] = s1[k][j] - cn1[j];
    		xyz[1][j] = s2[k][j] - cn2[j];
	}
    	for (j = 0; j< DIM3; j++)
    		for (i = 0; i< DIM3; i++)
		    tns[i][j] += xyz[0][j] * xyz[1][i];     
  }
}
/* -------------------------------------------------------------------- */
/* calculates the matrix product c = a * mat where a is a rotation	*/
/* around axis X, Y or Z. 						*/
/* -------------------------------------------------------------------- */
static void rot_xyz(double cos, double sin, double mat[3][3], int b, int c)
{
  	double 	arg1, arg2;
	int 	i;

	for (i = 0; i < DIM3; i++){
  		arg1 = mat[b][i];
  		arg2 = mat[c][i];
  		mat[b][i] = cos * arg1 - sin * arg2;
  		mat[c][i] = sin * arg1 + cos * arg2;
	}

  return;
}
/* -------------------------------------------------------------------- */
/* calculates angle and determines second derivative			*/
/*   tan(a) = +/- sin(a) / cos(a)					*/
/*   a = +/- atan(sin(a)/cos(a)) = atan(+/-sin(a)/cos(a))		*/
/*   if cos(a) * arg2 - sin(a) * arg1 > 0  solution is a minimum 	*/
/* -------------------------------------------------------------------- */
double rms_angle(double arg1, double arg2, double cs[2])
{
  double angle = 0.707;

	if (arg2 != 0.0)
  		angle = atan2(arg1, arg2);
	
  	cs[0] = cos(angle); 
	cs[1] = sin(angle);
  	if   (cs[0] * arg2 >= cs[1]* arg1) 
		return angle;
  	else { 
		cs[1]= -cs[1]; 
		return -angle;
	}
}
/* -------------------------------------------------------------------- */
/* find optimal rotations */
/* -------------------------------------------------------------------- */
double rms_cycle(double ts[3][3])
{
  double 	abs_a, 
		cs[2],
		sum = 0.0;                   /* sum of the 3 rotation angles */

  /* Find optimal rotation around x-axis */
  abs_a = fabs(rms_angle(ts[1][2]-ts[2][1], ts[1][1]+ts[2][2], cs));
  if (abs_a > rms_no_rot) {
  	rot_xyz(cs[0], cs[1], ts, 1, 2);
  	rot_xyz(cs[0], cs[1], rms_brm, 1, 2);
  }
  sum += abs_a;

  /* Find optimal rotation around y-axis */
  abs_a = fabs(rms_angle(ts[0][2]-ts[2][0], ts[0][0]+ts[2][2], cs));
  if (abs_a > rms_no_rot) {
  	rot_xyz(cs[0], cs[1], ts, 0, 2);
  	rot_xyz(cs[0], cs[1], rms_brm, 0, 2);
  }
  sum += abs_a;

  /* Find optimal rotation around z-axis */
  abs_a = fabs(rms_angle(ts[0][1]-ts[1][0], ts[0][0]+ts[1][1], cs));
  if (abs_a > rms_no_rot) {
  	rot_xyz(cs[0], cs[1], ts, 0, 1);
  	rot_xyz(cs[0], cs[1], rms_brm, 0, 1);
  }
  sum += abs_a;

return sum;
}
/* ------------------------------------------------------------------- */
/* calculates the matrix "rms_brm" of optimal superposition.
   returns i if iteration converged
   returns 0 if iteration did not converge  */
/* ------------------------------------------------------------------- */
int rms_optimat(double **s1, double **s2, double *cn1, double *cn2, int len)
{
	int	i;
	double tns[3][3];        /* cross tensor */

  	coo_unit_matrix(rms_brm);
  	rms_cross_tensor(s1, s2, tns, cn1, cn2, len);

  	for (i = 0; i < 100; i++) {
    		if (rms_cycle(tns) < rms_tolerance) {
			return i;
		}
  	}
  return 0;
}
/* -------------------------------------------------------------------- */
/* calcul de l'e.q.m et de la matrice de superposition 			*/
/* -------------------------------------------------------------------- */
double rap_eqm(double **p1, double **b2, int ll, double *cn1, double *cn2)
{

	rms_optimat(p1, b2, cn1, cn2, ll);
/*if (rms_error(p1, b2, cn1, cn2, ll) < 0.5)
fprintf(stdout,"opt:%5d %5.3f\n", rms_optimat(p1, b2, cn1, cn2, ll), rms_error(p1, b2, cn1, cn2, ll));
*/
  return rms_error(p1, b2, cn1, cn2, ll);
}
/*--------------------------------------------------------------------- */
/*  calculates geometric center s from coordinate array c of length n.	*/
/*   returns coordinates of center in s. 				*/
/*--------------------------------------------------------------------- */
double *coo_geo_center(double **s, double *c, int n)
{
	int 	i, j;
	double	   inv = 1./n;

	c[0] = c[1] = c[2] = 0.0;

 	for (j = 0; j< DIM3; j++){
		for (i = 0; i < n; i++) {
			c[j] += s[i][j];
	   	}
    		c[j] *= inv;
    	}

    return c;
}
/* ------------------------------------------------------------------ */
/* translates coordinates to center c					*/
/* ------------------------------------------------------------------ */
void move_coords(double **co, int n, double *c)
{
	int	i, j;

 	for (j = 0; j< DIM3; j++)
		for (i = 0; i < n; i++) {
	    		co[i][j] -= c[j];
  		}
}
/* -------------------------------------------------------------------- */
/* calculates geometric center from coordinate array coo of length n
   and translates coordinates to center. */
/* -------------------------------------------------------------------- */
void coo_center_coords(double **coo, int n)
{
  	int    i, j;
	double	   cxyz[3] = {0. ,0. ,0.}, inv = 1./n;

  	for (j = 0; j< DIM3; j++){
		for (i = 0; i < n; i++) 
			cxyz[j] = coo[i][j];	
		cxyz[j] *= inv;	
  		for (i = 0; i < n; i++) 
			coo[i][j] -= cxyz[j];	
  	}
}
/*--------------------------------------------------------------*/
/* END OF EQM.C							*/
/*--------------------------------------------------------------*/
