/* produit les calculs sur les coordonnees necessaires dans tito */
#include <stdio.h>
#include <stdlib.h>     /* contient malloc sur HP */
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "TITO/tito.h"
#include "TITO/macro2.h"

/* -------------------- */
/* parametres           */
/* -------------------- */
#define DIM3  3
#define XX  0
#define YY  1
#define ZZ  2

#define cos45 0.707106781
#define sin45 0.707106781

#define RADIANS 0.017453292

/* -------------------- */
/* variables            */
/* -------------------- */

extern	double	rms_brm[DIM3][DIM3];

/* -------------------------------------------------------------------- */
/* translates b to or from pmf position of a                            */
/* -------------------------------------------------------------------- */
double *mv_a_b(double *a, double *b, double r)
{
        double  r1 = 1-r;

        b[XX] = r*b[XX]+(r1)*a[XX];
        b[YY] = r*b[YY]+(r1)*a[YY];
        b[ZZ] = r*b[ZZ]+(r1)*a[ZZ];

 return b;
}
/* -------------------------------------------------------------------- */
/* translates b to or from pmf position of a                            */
/* -------------------------------------------------------------------- */
double *translate(double ***t, int i, double r, int b, int a)
{
        double  d, r1;

        r1 = 1 - r;
        t[i][b][XX] = r*t[i][b][XX]+(r1)*t[i][a][XX];
        t[i][b][YY] = r*t[i][b][YY]+(r1)*t[i][a][YY];
        t[i][b][ZZ] = r*t[i][b][ZZ]+(r1)*t[i][a][ZZ];

 return t[i][b];
}
/************************************************************************/
/* double  TEMP                                                         */
/************************************************************************/
void    re_cp_coords(double *tps, double *bc, double *c)
{

     	tps[XX] = bc[XX] + c[XX];
     	tps[YY] = bc[YY] + c[YY];
     	tps[ZZ] = bc[ZZ] + c[ZZ];
}
/************************************************************************/
/* double  TEMP                                                       */
/************************************************************************/
void    cp_mv_coords(double *b, double *c, double *cn2)
{
	b[XX] = c[XX] - cn2[XX];
	b[YY] = c[YY] - cn2[YY];
	b[ZZ] = c[ZZ] - cn2[ZZ];
/*	fprintf(stdout, "cn2 = %f %f %f\n", cn2[XX], cn2[YY], cn2[ZZ]);*/
}
/************************************************************************/
/* double  TEMP                                                       */
/************************************************************************/
void    rot_coords(double *b, double mat[DIM3][DIM3])
{
  int   i, j;
  double xyz[DIM3];

        for (j = 0; j< DIM3; j++)
                xyz[j] = b[j];
        for (j = 0; j< DIM3; j++)
        	b[j] =  mat[j][0] * xyz[0] + mat[j][1] * xyz[1] + mat[j][2] * xyz[2];
}
/********************************************************/
/*      recherche                                       */
/********************************************************/
void    mouline(double *bb, double *pp, double *from, double *cn1, double *cn2)
{

        cp_mv_coords(bb, from, cn2);
        rot_coords(bb, rms_brm);
	/*fprintf(stdout, "bb = %f %f %f\n", bb[XX], bb[YY], bb[ZZ]);*/
        re_cp_coords(pp, bb, cn1);
	/*fprintf(stdout, "pp = %f %f %f\n", pp[XX], pp[YY], pp[ZZ]);*/

}
/************************************************************************/
/* rebuild from internal coordinates                                    */
/************************************************************************/
void atom_pos(double *t, double *t1, double *t2, double *t3, double xn, double yn, double zn)
{
        double  ax,ay,az,bx,by,bz,cx,cy,cz,a,b,c,pm;

        ax = t2[XX] - t1[XX];           /* calcul des composantes vectorielles*/
        bx = t2[YY] - t1[YY];
        cx = t2[ZZ] - t1[ZZ];


        a = t3[XX] - t2[XX];            /* calcul des composantes vectorielles */
        b = t3[YY] - t2[YY];
        c = t3[ZZ] - t2[ZZ];

        az = bx*c - cx*b; bz = cx*a - ax*c; cz = ax*b - bx*a;

        zn *= PM(az, bz, cz);

        ay = bz*cx - cz*bx; by = cz*ax - az*cx; cy = az*bx - bz*ax;

        yn *= PM(ay, by, cy);

        t[XX] = ax*xn+ay*yn+az*zn + t1[XX];
        t[YY] = bx*xn+by*yn+bz*zn + t1[YY];
        t[ZZ] = cx*xn+cy*yn+cz*zn + t1[ZZ];

}
/* -------------------------------------------------------------------- */
/* calcul des angles diedres 		                                */
/* -------------------------------------------------------------------- */
double  atom_dihed(double *t, double *t1, double *t2, double *t3)
{
        double  a[DIM3], px1, py1, pz1,
                b[DIM3], px2, py2, pz2,
                c[DIM3], px3, py3, pz3,
                np1np2, dihed;
        int     i;

        for (i = 0; i < DIM3; i++){
                a[i] = t1[i] - t[i];
                b[i] = t2[i] - t1[i];
                c[i] = t3[i] - t2[i];
        }

        px1 =   a[YY]*b[ZZ] - b[YY]*a[ZZ];
        py1 = - a[XX]*b[ZZ] + b[XX]*a[ZZ];
        pz1 =   a[XX]*b[YY] - b[XX]*a[YY];

        px2 =   b[YY]*c[ZZ] - c[YY]*b[ZZ];
        py2 = - b[XX]*c[ZZ] + c[XX]*b[ZZ];
        pz2 =   b[XX]*c[YY] - c[XX]*b[YY];

        np1np2 = sqrt((px1*px1+py1*py1+pz1*pz1) * (px2*px2+py2*py2+pz2*pz2));

        if (np1np2 == 0)
                return -90*D2R;

        dihed = acos((px1*px2+py1*py2+pz1*pz2)/np1np2);

        px3 =   py1*pz2 - py2*pz1;
        py3 = - px1*pz2 + px2*pz1;
        pz3 =   px1*py2 - px2*py1;
        if ((px3*b[XX] + py3*b[YY] + pz3*b[ZZ]) <  0)            /* sine is < 0  */
                dihed = -dihed;


        return dihed;                     /* radian    */
}
/* -------------------------------------------------------------------- */
/* computes the angle defined by the 3 points				*/
/* given in ra. returns angle in [-PI, PI] rad.			*/
/* -------------------------------------------------------------------- */
double atom_angle(double *t, double *t1, double *t2)
{
	double	xa, ya, za, xb, yb, zb,
		xc, yc, zc, np1np2, ang;

				/* calcul des composantes vectorielles	*/

	/*for (i = 0; i < DIM3; i++){
		a[i] = t1[i] - t[i];
		b[i] = t2[i] - t1[i];
	}
	xa = ra->x - (ra+1)->x;
	ya = ra->y - (ra+1)->y;
	za = ra->z - (ra+1)->z;
	xb = (ra+2)->x - (ra+1)->x;
	yb = (ra+2)->y - (ra+1)->y;
	zb = (ra+2)->z - (ra+1)->z;*/

	xa = t1[XX] - t[XX];
	ya = t1[YY] - t[YY];
	za = t1[ZZ] - t[ZZ];

	xb = t2[XX] - t1[XX];
	yb = t2[YY] - t1[YY];
	zb = t2[ZZ] - t1[ZZ];
	
	np1np2 = sqrt((xa*xa+ya*ya+za*za) * (xb*xb+yb*yb+zb*zb)); /* (produit des normes)			*/

	if (np1np2 == 0) 
			return NO_ANGLE*D2R;

	xc =   ya*zb - yb*za;
	yc = - xa*zb + xb*za;		/* produit vectoriel 1*2		*/
	zc =   xa*yb - xb*ya;
				
	ang = asin(sqrt(xc*xc+yc*yc+zc*zc)/np1np2); /* arcsinus de norme du produit vect.	*/

				/* ang is between -Pi/2 and Pi/2	*/
				/* we need to know the sign		*/
				/* of cosinus (produit scalaire)	*/
	
	if ((xa*xb+ya*yb+za*zb) >  0)			/* cos is < 0	*/
		ang = PI - ang;

	return ang;				/* rad 	*/

}
/****************************************************************************************/
/*	test d'assignement 								*/
/****************************************************************************************/
int	test_consa(float **c, int n)
{

        return ( (TV(c[n][ALFA], 25., 75.) && TV(c[n][TAU], 79, 100.))
               ||  (TV(c[n][ALFA], 38., 64.) && TV(c[n][TAU], 77, 106.))
               || (TV(c[n][D4], 4.75, 5.55) && TV(c[n][D5], 5.75, 6.45))
               );

}
/****************************************************************************************/
/*	test d'assignement 								*/
/****************************************************************************************/
int	test_consb(float **c, int n)
{

        return (
                ((TV(c[n][ALFA], -180.0, -150.) || TV(c[n][ALFA], 165., 180.0) )
                && TV(c[n][TAU], 105., 145.))
                ||
                ((TV(c[n][ALFA], -180.0, -110.) || TV(c[n][ALFA], 140., 180.0) )
                && TV(c[n][TAU], 120., 146.))
                || (TV(c[n][D3], 9.00, 10.90) && TV(c[n][D2], 6., 7.5)));
}
/****************************************************************************************/
/*	detection de feuillet		 						*/
/****************************************************************************************/
int	test_feuillet(float **c, int n, double **ATOM, int *HP[N_ATOM], int nres)
{
	int 	SBETA = 0, i, j;

	SBETA = 0;
	for (i = 0; i < 3 && n+i < nres; i++){
		for (j = 0; j < nres && j < n-2; j++) {
			if (!(TV(c[j][TAU], 117., 145.) || test_consa(c, j)))
				continue;
			if (TV(SQR_DIS(ATOM[HP[CA][n+i]], ATOM[HP[CA][j]]), 4.25, 5.25))
				SBETA++;
 		}
		for (j = n+3; j < nres; j++) {
			if ( !(TV(c[j][TAU], 117., 145.) || test_consa(c, j)))
				continue;
			if (TV(SQR_DIS(ATOM[HP[CA][n+i]], ATOM[HP[CA][j]]), 4.25, 5.25))
				SBETA++;
		}
	}
	return (SBETA > 4);
}
/****************************************************************************************/
/*	assignement 									*/
/****************************************************************************************/
float   **analyse_coor_int(float **coin, char **srot, double **ATOM, int *HP[N_ATOM], int nres)
{
	int	n = 0, p = 0, j;

        if (!(coin = NEWN(nres, float *))) return NULL;
        for (n = 0; n < nres; n++)
        	if (!(coin[n] = NEWN(N_COIN, float)))
			return NULL;
	for (n = 0; n < nres; n++) {
		srot[CONS][n] = COIL;
	}
	for (n = 1; n < nres-2; n++) {
		coin[n][D2] = sqrt(SQR_DIS(ATOM[HP[CA][n-1]],ATOM[HP[CA][n+1]]));
		coin[n][D3] = sqrt(SQR_DIS(ATOM[HP[CA][n-1]],ATOM[HP[CA][n+2]]));
		coin[n][TAU] = R2D*atom_angle(ATOM[HP[CA][n-1]], ATOM[HP[CA][n]], ATOM[HP[CA][n+1]]);
		/*	fprintf(stdout, "%d %5.1f\n", n, coin[n][TAU]);*/
		if (n < nres-3){
			coin[n][D4] = sqrt(SQR_DIS(ATOM[HP[CA][n-1]],ATOM[HP[CA][n+3]]));
			coin[n][ALFA] = R2D*atom_dihed(ATOM[HP[CA][n-1]], ATOM[HP[CA][n]], ATOM[HP[CA][n+1]], ATOM[HP[CA][n+2]]);
		}
		if (n < nres-4)
			coin[n][D5] = sqrt(SQR_DIS(ATOM[HP[CA][n-1]],ATOM[HP[CA][n+4]]));
	}
return coin;
}
/****************************************************************************************/
/*	assignement 									*/
/****************************************************************************************/
void   analyse_cons234(float **coin, char **srot, double **ATOM, int *HP[N_ATOM], int nres)
{
	char	beta = 'b'-'a', alpha = 'a'-'a', coil = 'c'-'a', bulge = 'u'-'a', turn = 't' - 'a', sheet = 's'-'a';
	float	nd, aai = 79., aas = 99.;
	int	n = 0, SBETA = 0, p = 0, j;

	coin = analyse_coor_int(coin, srot, ATOM, HP, nres);
/*	fprintf(stdout, "\n\tAssignation des helices\n");*/
	for (j = 1; j < nres-2; j++) {
		for (n = j; n < nres-2 && test_consa(coin,n); n++) ;
		if (n < nres-1)
			if (TV(coin[n][D4], 4.80, 5.45) && TV(coin[n][TAU], aai, aas))
				n++;
		if (n-j > 4) {
			if (TV(coin[j-1][D4], 4.75, 5.45) && TV(coin[j-1][TAU], aai, aas))
				srot[CONS][j-1] = ALFA;
			for (; j < nres-2 && j < n; j++)
				srot[CONS][j] = ALFA;                   
			if (TV(coin[j][D4], 4.75, 5.45) && TV(coin[j][TAU], aai, aas))
				srot[CONS][j] = ALFA;
		}
	}
/*	fprintf(stdout, "\n\tAssignation des brins\n");*/
	for (j = 1; j < nres-2; j++) {
		if (srot[CONS][j] == ALFA)
			continue;
		for (n = j; n < nres-2 && test_consb(coin,n); n++) ;
		if (n < nres-1)
			if (TV(coin[n][D2], 6.1, 7.3) && TV(coin[n][TAU], 117., 149.))
				n++;
		if (n-j > 3 || (n-j == 3 && test_feuillet(coin, n, ATOM, HP, nres))) {
			if (TV(coin[j-1][D2], 6.35, 7.35) )
				srot[CONS][j-1] = BETA;
   		for (; j < nres-2 && j < n; j++)
				srot[CONS][j] = BETA;                   
			if (TV(coin[j][D2], 6.35, 7.30)  && TV(coin[j][TAU], 118., 145.))
				srot[CONS][j] = BETA;
		}
	}
	/*for (n = 0; n < nres; n++) {
		fprintf(stdout, "%d %c\n", n, srot[CONS][n]+'a');
	}*/

	/*fprintf(stdout, "\n\tAssignation des segments\n");*/
/*-------------------------------------------------------------------------------------*/
/*	for (res = first_res->next; res->next; res = res->next) {
		if (res->cons == beta)
			continue;
		if (!test_consa(res) && !test_consa(res->next)) {
			if (TV(res->dis13, 5.0, 5.8) && TV(res->prev->dis13, 5.0, 5.8)
					&& TV(res->prev->dis14, 4.3, 5.3)) {
					res->cons = turn;
					res->next->cons = turn;
			}
		}
	}*/
/*-------------------------------------------------------------------------------------*/
			/*if (TV(res->dis13, 5.0, 6.0) && TV(res->prev->dis13, 5.0, 6.0)
					&& TV(res->prev->dis14, 5.5, 6.5)){
					res->cons = bulge;
					res->next->cons = bulge;
			}*/
}
/*-------------------------------------------------------------------------------------*/
/* The End */
/*-------------------------------------------------------------------------------------*/
